# Generated by Django 5.0.14 on 2025-07-16 12:39

from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0080_alter_trade_drift_factor_alter_trade_weighting'),
    ]

    operations = [
        migrations.AlterField(
            model_name='trade',
            name='drift_factor',
            field=models.DecimalField(decimal_places=16, default=Decimal('1'), help_text='Drift factor to be applied to the previous portfolio weight to get the actual effective weight including daily return', max_digits=19, verbose_name='Drift Factor'),
        ),
    ]
